#include <math.h>
#include <stdio.h>
#include "sizes.h"
#include "algo.h"
#include "def.h"


/* Remap picture from type 2 targa (B,G,R) to 8bpp
 */

void RemapPicture(ubyte *TgaBas,ubyte *dest,ubyte ICMap[],full count)
{
	full a;
	
	for (count-=1;count>=0;count--)
	{
		a=0;
		a|=*TgaBas++>>2;
		a|=(ufull)(*TgaBas++&0xFC)<<4;
		a|=(ufull)(*TgaBas++&0xFC)<<10;
		
		*dest++=ICMap[a];
	}
}


/* Make Inverse Colour Map
 */

void MakeICMap(BOX BoxBase[],ubyte cm[],ubyte icm[], ufull hist[], half nbox)
{
	double MaxLen,MinLen,len,z=(double)255/63;
	ubyte *cmp,col;
	full rm,gm,bm,r,g,b,r2,g2,b2,icmi;
	ufull *rp,*gp,*bp;
	half i,j,*CBase,*CTabP,A[COLOURS-1],AC,*AP,hindex,
	     CTab[8*3]={0,1,2, 0,1,5, 0,4,2, 0,4,5,
	                3,1,2, 3,1,5, 3,4,2, 3,4,5};
	
	for (i=nbox-1;i>=0;i--)
	{
		AC=0;
		MaxLen=0;
		CBase=&BoxBase[i].r1;
		CTabP=CTab;
		cmp=&cm[i*3];
		rm=*cmp++;
		gm=*cmp++;
		bm=*cmp++;
		
		for (j=7;j>=0;j--)
		{
			r=(full)round(CBase[*CTabP++]*z);
			g=(full)round(CBase[*CTabP++]*z);
			b=(full)round(CBase[*CTabP++]*z);
			
			len=sqrt((double)((r-rm)*(r-rm)+(g-gm)*(g-gm)+(b-bm)*(b-bm)));
			if (len>MaxLen)
				MaxLen=len;
		}
		
		MaxLen*=2;
		
		for (j=nbox-1;j>=0;j--)
		{
			if (j!=i)
			{
				cmp=&cm[j*3];
				r=*cmp++;
				g=*cmp++;
				b=*cmp++;
				
				len=sqrt((double)((r-rm)*(r-rm)+(g-gm)*(g-gm)+(b-bm)*(b-bm)));
				if (len<MaxLen)
					A[AC++]=j;
			}
		}
		
		rp=hist + CBase[0]*R_STRIDE + CBase[1]*G_STRIDE + CBase[2];
		
		for (r2=CBase[0];r2<=CBase[3];r2++)
		{
			gp=rp;
			
			for (g2=CBase[1];g2<=CBase[4];g2++)
			{
				bp=gp;
				
				for (b2=CBase[2];b2<=CBase[5];b2++)
				{
					if (*bp++)
					{
						icmi=(r2<<12)|(g2<<6)|b2;
						
						if (!AC)
							icm[icmi]=(ubyte)i;
						else
						{
							r=(full)round(r2*z);
							g=(full)round(g2*z);
							b=(full)round(b2*z);
							
							MinLen=sqrt((double)((r-rm)*(r-rm)+(g-gm)*(g-gm)+(b-bm)*(b-bm)));
							col=i;
							
							AP=A;
							for (AC--;AC>=0;AC--)
							{
								hindex=*AP++;
								cmp=&cm[hindex*3];
								rm=*cmp++;
								gm=*cmp++;
								bm=*cmp++;
								
								len=sqrt((double)((r-rm)*(r-rm)+(g-gm)*(g-gm)+(b-bm)*(b-bm)));
								if (len<MinLen)
								{
									MinLen=len;
									col=hindex;
								}
							}
							
							icm[icmi]=(ubyte)col;
						}
					}
					
				}
				gp+=G_STRIDE;
			}
			rp+=R_STRIDE;
		}
		
	}
}


/* Make colourmap
 */

void MakeColourMap(BOX b[],ubyte cm[],half n)
{
	double k=(double)255/63;
	
	for (n-=1;n>=0;n--)
	{
		*cm++=round((b->t_ur*k)/b->tw);
		*cm++=round((b->t_ug*k)/b->tw);
		*cm++=round((b->t_ub*k)/b->tw);
		b++;
	}
}


/* 18 bit histogram
 * input: type 2 targa (B,G,R)
 */

void MakeHistogram(full count,ubyte *TgaBas,ufull *hist)
{
	ufull a,*ufp=hist;
	
	for (a=0;a<HSIZE/4;a++)
		*ufp++=0;
	
	for (count-=1;count>=0;count--)
	{
		a=0;
		a|=*TgaBas++>>2;
		a|=(ufull)(*TgaBas++&0xFC)<<4;
		a|=(ufull)(*TgaBas++&0xFC)<<10;
		(*(hist+a))++;
	}
}


/* Variance calculation
 */

ufull Variance(STAT *s)
{
	double a;
	
	a=(double)s->t_ur*(double)s->t_ur;
	a+=(double)s->t_ug*(double)s->t_ug;
	a+=(double)s->t_ub*(double)s->t_ub;
	a/=(double)s->tw;
	
	return ((ufull)((double)s->tt_sum-a));
}


/* Gathering statistics from a
 * colour space box
 */

void Sum(STAT *s, COORD *c, ufull *hist){
	
	half r,g,b;
	ufull i=0,j=0,rs,ts,w,tr=0,tg=0,tb=0,*rp,*gp,*bp;
	
	rp=hist + c->r1*R_STRIDE + c->g1*G_STRIDE + c->b1;
	
	for (r=c->r1; r<=c->r2; r++)
	{
		rs=r*r;
		gp=rp;
		
		for (g=c->g1; g<=c->g2; g++)
		{
			ts=rs+g*g;
			bp=gp;
			
			for (b=c->b1; b<=c->b2; b++)
			{
				w=*bp++;
				if (w)
				{
					j+=w;
					tr+=r*w;
					tg+=g*w;
					tb+=b*w;
					i+=(ts+b*b)*w;
				}
			}
			gp+=G_STRIDE;
		}
		rp+=R_STRIDE;
	}
	s->tw=j;
	s->tt_sum=i;
	s->t_ur=tr;
	s->t_ug=tg;
	s->t_ub=tb;
}


/* Shrink box function
 */

void ShrinkBox(COORD *c, ufull *hist)
{
	half safe,i,j,a=1,b=3,*cp=(half *)c;
	
	for (i=1;i>=0;i--)
	{
		for (j=2;j>=0;j--)
		{
			safe=*(cp+b);
			*(cp+b)=*cp;
			
			while (SpaceUsed(c,hist)==0L)
 				*(cp+b)=*cp+=a;
			
			*(cp+b)=safe;
			cp++;
		}
		a=-a;b=-b;
	}
}


/* Returns weight in histogram space
 */

ufull SpaceUsed(COORD *c, ufull *hist)
{
	half r,g,b;
	ufull w=0,*rp,*gp,*bp;
	
	rp=hist + c->r1*R_STRIDE + c->g1*G_STRIDE + c->b1;
	
	for (r=c->r2; r>=c->r1; r--)
	{
		gp=rp;
		
		for (g=c->g2; g>=c->g1; g--)
		{
			bp=gp;
			
			for (b=c->b2; b>=c->b1; b--)
				w+=*bp++;
			
			gp+=G_STRIDE;
		}
		rp+=R_STRIDE;
	}
	return w;
}


/* Splits a colourbox at the best place
 */

int SplitBox(BOX *FirstBox,half nbox,ufull *hist)
{
	half r,g,b,i,j,c[6],f;
	ufull rv=0xffffffffL,gv=rv,bv=rv,tv=0,uftmp,
	      s1[6],s2[6],s3[6];
	BOX *b1=FirstBox,*b2=FirstBox+nbox,*bpek;
	
	/* Find box to split
	 * (Probably not the best method)
	 */
	for (i=nbox-1;i>=0;i--)
	{
		uftmp=b1->v;
		if (uftmp>tv)
		{
			tv=uftmp;
			bpek=b1;
		}
		b1++;
	}
	
	if (tv==0)
		f=1;
	else
	{
		b1=bpek;
		
		/* Red axis
		 * Do some initializations
		 */
		for (i=5;i>=0;i--)
		{
			s1[i]=0;
			s2[i]=*(&b1->v+i);
			c[i]=*(&b1->r1+i);
		}
		
		/* Locate optimum splitpoint on red axis */
		for (i=b1->r1;i<b1->r2;i++)
		{
			/* Calculate the statistics for the area being taken
			 * away from the first box and given to the left box
			 */
			c[0]=c[3]=i;
			Sum((STAT *)s3,(COORD *)c,hist);
			for (j=5;j>0;j--)
			{
				s1[j]+=s3[j];
				s2[j]-=s3[j];
			}
			
			/* Calculate the first and second box's overall variance */
			tv=Variance((STAT *)s1)+Variance((STAT *)s2);
			
			/* Found a better split point? if so remember it */
			if (tv<rv)
			{
				rv=tv;
				r=i;
			}
		}
		
		/* Green axis */
		for (i=5;i>=0;i--)
		{
			s1[i]=0;
			s2[i]=*(&b1->v+i);
			c[i]=*(&b1->r1+i);
		}
		
		for (i=b1->g1;i<b1->g2;i++)
		{
			c[1]=c[4]=i;
			Sum((STAT *)s3,(COORD *)c,hist);
			for (j=5;j>0;j--)
			{
				s1[j]+=s3[j];
				s2[j]-=s3[j];
			}
			tv=Variance((STAT *)s1)+Variance((STAT *)s2);
			if (tv<gv)
			{
				gv=tv;
				g=i;
			}
		}
	
		/* Blue axis */
		for (i=5;i>=0;i--)
		{
			s1[i]=0;
			s2[i]=*(&b1->v+i);
			c[i]=*(&b1->r1+i);
		}
		
		for (i=b1->b1;i<b1->b2;i++)
		{
			c[2]=c[5]=i;
			Sum((STAT *)s3,(COORD *)c,hist);
			for (j=5;j>0;j--)
			{
				s1[j]+=s3[j];
				s2[j]-=s3[j];
			}
			tv=Variance((STAT *)s1)+Variance((STAT *)s2);
			if (tv<bv)
			{
				bv=tv;
				b=i;
			}
		}
		
		/* Copy the original coordinates to the second box */
		for (i=5;i>=0;i--)
			*(&b2->r1+i)=*(&b1->r1+i);
		
		/* Find out which axis should be split
		 * (Intensity=.30R+.59G+.11B, therefore green first)
		 */
		tv=gv;i=1;
		if (rv<tv)
		{
			tv=rv;i=0;
		}
		if (bv<tv)
			i=2;
		
		switch (i)
		{
			case 0:
			{
				b1->r2=r;
				b2->r1=r+1;
				break;
			}
			case 1:
			{
				b1->g2=g;
				b2->g1=g+1;
				break;
			}
			case 2:
			{
				b1->b2=b;
				b2->b1=b+1;
			}
		}
		
		/* Shrink the new boxes to their minimum possible sizes */
		ShrinkBox((COORD *)&b1->r1,hist);
		ShrinkBox((COORD *)&b2->r1,hist);
		
		/* Update statistics */
		Sum((STAT *)b2,(COORD *)&b2->r1,hist);
		for (i=4;i>=0;i--)
			*(&b1->tw+i)-=*(&b2->tw+i);
		
		b1->v=Variance((STAT *)b1);
		b2->v=Variance((STAT *)b2);
		
		f=0;
	}
	
	return f;
}